#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from abc import ABC, abstractmethod


from apipeline.frames.base import Frame


class FrameSerializer(ABC):
    @abstractmethod
    def serialize(self, frame: Frame) -> str | bytes | None:
        pass

    @abstractmethod
    def deserialize(self, data: str | bytes) -> Frame | None:
        pass



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from typing import Coroutine

from apipeline.frames.data_frames import Frame, TextFrame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class StatelessTextTransformer(FrameProcessor):
    """This processor calls the given function on any text in a text frame.

    >>> async def print_frames(aggregator, frame):
    ...     async for frame in aggregator.process_frame(frame):
    ...         print(frame.text)

    >>> aggregator = StatelessTextTransformer(lambda x: x.upper())
    >>> asyncio.run(print_frames(aggregator, TextFrame("Hello")))
    HELLO
    """

    def __init__(self, transform_fn):
        super().__init__()
        self._transform_fn = transform_fn

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, TextFrame):
            result = self._transform_fn(frame.text)
            if isinstance(result, Coroutine):
                result = await result
            await self.push_frame(TextFrame(text=result))
        else:
            await self.push_frame(frame, direction)



import logging
import asyncio
from abc import ABC, abstractmethod

from apipeline.frames.base import Frame
from apipeline.frames.data_frames import DataFrame
from apipeline.frames.sys_frames import CancelFrame, MetricsFrame, StartInterruptionFrame, StopInterruptionFrame, SystemFrame
from apipeline.frames.control_frames import ControlFrame, EndFrame, StartFrame
from apipeline.processors.async_frame_processor import AsyncFrameProcessor
from apipeline.processors.frame_processor import FrameDirection


class OutputProcessor(AsyncFrameProcessor, ABC):
    r"""
    output processor with start, stop control frames and cancel sys frames
    sink data frame and control frames
    """

    def __init__(
            self,
            *,
            name: str | None = None,
            loop: asyncio.AbstractEventLoop | None = None,
            **kwargs):
        super().__init__(name=name, loop=loop, **kwargs)

        self._stopped_event = asyncio.Event()
        self._sink_event = asyncio.Event()

        # Create sink frame task. This is the task that will actually write
        # audio or video frames. We write audio/video in a task so we can keep
        # generating frames upstream while, for example, the audio is playing.
        self._create_sink_task()

    async def start(self, frame: StartFrame):
        if self._sink_task.cancelled():
            self._create_sink_task()

    async def stop(self, frame: EndFrame):
        # Wait for the push frame and sink tasks to finish. They will finish when
        # the EndFrame is actually processed.
        await self._push_frame_task
        await self._sink_task

    async def cancel(self, frame: CancelFrame):
        # Cancel all the tasks and wait for them to finish.
        self._push_frame_task.cancel()
        await self._push_frame_task

        self._sink_task.cancel()
        await self._sink_task

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        #
        # System frames (like StartInterruptionFrame) are pushed immediately.
        # Other frames require order so they are put in the sink queue.
        #
        if isinstance(frame, CancelFrame):
            await self.push_frame(frame, direction)
            await self.cancel(frame)
        elif isinstance(frame, StartInterruptionFrame) or isinstance(frame, StopInterruptionFrame):
            await self.push_frame(frame, direction)
            await self._handle_interruptions(frame)
        elif isinstance(frame, MetricsFrame):
            await self.push_frame(frame, direction)
            await self.send_metrics(frame)
        elif isinstance(frame, SystemFrame):
            await self.push_frame(frame, direction)
        # Control frames.
        elif isinstance(frame, StartFrame):
            await self._sink_queue.put(frame)
            await self.start(frame)
        elif isinstance(frame, EndFrame):
            await self._sink_queue.put(frame)
            await self.stop(frame)
        else:
            await self._sink_queue.put(frame)

    async def send_metrics(self, frame: MetricsFrame):
        pass

    async def _handle_interruptions(self, frame: Frame):
        if not self.interruptions_allowed:
            return

        if isinstance(frame, StartInterruptionFrame):
            # Stop sink task.
            self._sink_task.cancel()
            await self._sink_task
            self._create_sink_task()
            # Stop push task.
            self._push_frame_task.cancel()
            await self._push_frame_task
            self._create_push_task()

    #
    # sink frames task
    #

    def _create_sink_task(self):
        loop = self.get_event_loop()
        self._sink_queue = asyncio.Queue()
        self._sink_task = loop.create_task(self._sink_task_handler())

    async def _sink_task_handler(self):
        running = True
        while running:
            try:
                frame = await self._sink_queue.get()
                # print(f"_sink_queue.get: {frame}")
                # sink data frame
                if isinstance(frame, DataFrame):
                    await self.sink(frame)
                    # subclass need wait sink task done
                    await self._sink_event.wait()
                    self._sink_event.clear()
                # sink control frame
                elif isinstance(frame, ControlFrame):
                    await self.sink_control_frame(frame)
                else:
                    await self.queue_frame(frame)

                running = not isinstance(frame, EndFrame)

                self._sink_queue.task_done()
            except asyncio.CancelledError:
                break
            except Exception as ex:
                logging.exception(f"{self} error processing sink queue: {ex}")

    @abstractmethod
    async def sink(self, frame: DataFrame):
        """
        Multimoding(text,audio,image) Sink, use _sink_event to set
        """
        raise NotImplementedError

    async def sink_control_frame(self, frame: ControlFrame):
        """
        flow stream to control frame start, end, e.g. until await task future is finished
        """
        logging.debug(
            f"{self.__class__.__name__} process_control_frame {frame} doing")
        await self.queue_frame(frame)


class OutputFrameProcessor(OutputProcessor):
    """
    sink data frames to asyncio.Queue
    if have callback, use out push task handler to consume the queue ;
    else user can get the out_queue to consume;
    """

    def __init__(self, *, cb=None, name: str | None = None,
                 loop: asyncio.AbstractEventLoop | None = None, **kwargs):
        super().__init__(name=name, loop=loop, **kwargs)
        self._out_queue = asyncio.Queue()
        self._cb = cb
        self._out_task = None
        if self._cb:
            self._running = True
            self._create_output_task()

    @property
    def out_queue(self):
        if self._cb:
            return None
        return self._out_queue

    def set_sink_event(self):
        if self._cb:
            return
        self._sink_event.set()

    async def sink(self, frame: DataFrame):
        await self._out_queue.put(frame)

    async def sink_control_frame(self, frame: ControlFrame):
        self._running = not isinstance(frame, EndFrame)
        return await super().sink_control_frame(frame)

    def _create_output_task(self):
        self._out_task = self.get_event_loop().create_task(self._out_push_task_handler())

    async def _out_push_task_handler(self):
        while self._running:
            try:
                frame = await asyncio.wait_for(self._out_queue.get(), 0.1)
                # print(f"_out_queue.get: {frame}")
                if asyncio.iscoroutinefunction(self._cb):
                    await self._cb(frame)
                else:
                    self._cb(frame)
                self._out_queue.task_done()
                self._sink_event.set()
            except TimeoutError:
                continue
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                logging.warning("Output push task was cancelled.")
                break
            except Exception as ex:
                logging.exception(f"Unexpected error in _out_push_task_handler: {ex}")

    async def start(self, frame: StartFrame):
        if self._out_task is None or self._out_task.cancelled():
            self._create_output_task()
        await super().start(frame)

    async def stop(self, frame: EndFrame):
        self._out_task and await self._out_task
        await super().stop(frame)

    async def cancel(self, frame: CancelFrame):
        if self._out_task and self._out_task.cancelled() is False:
            self._out_task.cancel()
            await self._out_task
            self._out_task = None
        await super().cancel(frame)



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import asyncio
import logging
import signal

from apipeline.pipeline.task import PipelineTask
from apipeline.utils.obj import obj_count, obj_id


class PipelineRunner:

    def __init__(self, name: str | None = None, handle_sigint: bool = True):
        self.id: int = obj_id()
        self.name: str = name or f"{self.__class__.__name__}#{obj_count(self)}"

        self._tasks = {}

        if handle_sigint:
            self._setup_sigint()

    async def run(self, task: PipelineTask):
        logging.debug(f"Runner {self} started running {task}")
        self._tasks[task.name] = task
        await task.run()
        del self._tasks[task.name]
        logging.debug(f"Runner {self} finished running {task}")

    async def stop_when_done(self):
        logging.debug(f"Runner {self} scheduled to stop when all tasks are done")
        await asyncio.gather(*[t.stop_when_done() for t in self._tasks.values()])

    async def cancel(self):
        logging.debug(f"Canceling runner {self}")
        await asyncio.gather(*[t.cancel() for t in self._tasks.values()])

    def _setup_sigint(self):
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(
            signal.SIGINT,
            lambda *args: asyncio.create_task(self._sig_handler())
        )
        loop.add_signal_handler(
            signal.SIGTERM,
            lambda *args: asyncio.create_task(self._sig_handler())
        )

    async def _sig_handler(self):
        logging.warning(f"Interruption detected. Canceling runner {self}")
        await self.cancel()

    def __str__(self):
        return self.name



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from typing import Awaitable, Callable
import asyncio


from apipeline.frames.sys_frames import Frame, StartInterruptionFrame, StopInterruptionFrame, SystemFrame
from apipeline.processors.async_frame_processor import AsyncFrameProcessor
from apipeline.processors.frame_processor import FrameDirection


class UserIdleProcessor(AsyncFrameProcessor):
    """This class is useful to check if the user is interacting with the bot
    within a given timeout. If the timeout is reached before any interaction
    occurred the provided callback will be called.

    The callback can then be used to push frames downstream by using
    `queue_frame()` (or `push_frame()` for system frames).

    """

    def __init__(
            self,
            *,
            callback: Callable[["UserIdleProcessor"], Awaitable[None]],
            timeout: float,
            **kwargs):
        super().__init__(**kwargs)

        self._callback = callback
        self._timeout = timeout

        self._interrupted = False

        self._create_idle_task()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, SystemFrame):
            await self.push_frame(frame, direction)
        else:
            await self.queue_frame(frame, direction)

    async def cleanup(self):
        self._idle_task.cancel()
        await self._idle_task

    def _create_idle_task(self):
        self._idle_event = asyncio.Event()
        self._idle_task = self.get_event_loop().create_task(self._idle_task_handler())

    async def _idle_task_handler(self):
        while True:
            try:
                await asyncio.wait_for(self._idle_event.wait(), timeout=self._timeout)
            except asyncio.TimeoutError:
                if not self._interrupted:
                    await self._callback(self)
            except asyncio.CancelledError:
                break
            finally:
                self._idle_event.clear()



from typing import Optional
import logging

from apipeline.frames.base import Frame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class FrameLogger(FrameProcessor):
    def __init__(
            self,
            prefix="Frame",
            color: Optional[str] = None,
            ignored_frame_types: Optional[list] = [],
            include_frame_types: Optional[list] = None):
        super().__init__()
        self._prefix = prefix
        self._color = color
        self._ignored_frame_types = tuple(
            ignored_frame_types) if ignored_frame_types is not None else None
        self._include_frame_types = tuple(include_frame_types) if include_frame_types else None

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        if self._ignored_frame_types is not None and not isinstance(
                frame, self._ignored_frame_types):
            if not self._include_frame_types \
                or (self._include_frame_types
                    and isinstance(frame, self._include_frame_types)):
                from_to = f"{self._prev} ---> {self}"
                if direction == FrameDirection.UPSTREAM:
                    from_to = f"{self} <--- {self._next} "
                msg = f"{from_to} {self._prefix}: {frame}"
                if self._color:
                    msg = f"<{self._color}>{msg}</>"
                logging.info(msg)

        await self.push_frame(frame, direction)



import unittest
from apipeline.frames.control_frames import EndFrame
from apipeline.frames.sys_frames import CancelFrame, StopTaskFrame
from apipeline.pipeline.pipeline import Pipeline, FrameDirection
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import Frame, TextFrame, ImageRawFrame, AudioRawFrame
from apipeline.processors.frame_processor import FrameProcessor


"""
python -m unittest tests.test_simple.TestSimple
"""


class FrameTraceLogger(FrameProcessor):

    async def process_frame(
            self,
            frame: Frame,
            direction: FrameDirection = FrameDirection.DOWNSTREAM):
        await super().process_frame(frame, direction)

        from_to = f"{self._prev} ---> {self}"
        if direction == FrameDirection.UPSTREAM:
            from_to = f"{self} <--- {self._next} "
        print(f"{from_to} get Frame: {frame}")


class TestSimple(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):

        pipeline = Pipeline([
            FrameTraceLogger(),
        ])

        self.task = PipelineTask(
            pipeline,
            PipelineParams()
        )

    async def asyncTearDown(self):
        pass

    async def test_run(self):
        runner = PipelineRunner()
        await self.task.queue_frame(TextFrame("你好"))
        await self.task.queue_frame(ImageRawFrame(image=bytes([]), size=(0, 0), format="PNG", mode="RGB"))
        await self.task.queue_frame(AudioRawFrame(audio=bytes([])))
        await self.task.queue_frame(EndFrame())
        await runner.run(self.task)



import asyncio
from typing import Type

from apipeline.frames import CancelFrame, EndFrame, Frame, StartFrame
from apipeline.processors.async_frame_processor import AsyncFrameProcessor
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.notifiers.base import BaseNotifier


class HoldFramesAggregator(AsyncFrameProcessor):
    """
    This aggregator keeps the specified frames,
    it doesn't let the frames through until the notifier is notified.
    """

    def __init__(self,
                 notifier: BaseNotifier,
                 hold_frame_classes: tuple[Type[Frame]],
                 **kwargs):
        super().__init__(**kwargs)
        self._notifier = notifier
        self._hold_frame_classes = hold_frame_classes
        self._hold_frames = []

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self.push_frame(frame)
            await self._start()
        if isinstance(frame, (EndFrame, CancelFrame)):
            await self.push_frame(frame, direction)
            await self._stop()
        elif isinstance(frame, self._hold_frame_classes):
            self._hold_frames.append(frame)
        else:
            await self.push_frame(frame, direction)

    async def _start(self):
        self._gate_task = self.get_event_loop().create_task(self._gate_task_handler())

    async def _stop(self):
        self._gate_task.cancel()
        await self._gate_task

    async def _gate_task_handler(self):
        while True:
            try:
                await self._notifier.wait()
                for frame in self._hold_frames:
                    await self.queue_frame(frame)
                self._hold_frames = []
            except asyncio.CancelledError:
                break


class HoldLastFrameAggregator(AsyncFrameProcessor):
    """
    This aggregator keeps the specified last frame,
    it doesn't let the frame through until the notifier is notified.
    """

    def __init__(self,
                 notifier: BaseNotifier,
                 hold_frame_classes: tuple[Type[Frame]],
                 **kwargs):
        super().__init__(**kwargs)
        self._notifier = notifier
        self._hold_frame_classes = hold_frame_classes
        self._hold_last_frame = None

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self.push_frame(frame)
            await self._start()
        if isinstance(frame, (EndFrame, CancelFrame)):
            await self._stop()
            await self.push_frame(frame, direction)
        elif isinstance(frame, self._hold_frame_classes):
            self._hold_last_frame = frame
        else:
            await self.push_frame(frame, direction)

    async def _start(self):
        self._gate_task = self.get_event_loop().create_task(self._gate_task_handler())

    async def _stop(self):
        self._gate_task.cancel()
        await self._gate_task

    async def _gate_task_handler(self):
        while True:
            try:
                await self._notifier.wait()
                if self._hold_last_frame:
                    await self.queue_frame(self._hold_last_frame)
                self._hold_last_frame = None
            except asyncio.CancelledError:
                break



from dataclasses import dataclass

from .base import Frame

#
# Control frames
#


@dataclass
class ControlFrame(Frame):
    pass


@dataclass
class StartFrame(ControlFrame):
    """This is the first frame that should be pushed down a pipeline."""
    allow_interruptions: bool = False
    enable_metrics: bool = False
    enable_usage_metrics: bool = False
    report_only_initial_ttfb: bool = False


@dataclass
class EndFrame(ControlFrame):
    """Indicates that a pipeline has ended and frame processors and pipelines
    should be shut down. If the transport receives this frame, it will stop
    sending frames to its output channel(s) and close all its threads. Note,
    that this is a control frame, which means it will received in the order it
    was sent (unline system frames).

    """
    pass


@dataclass
class SyncFrame(ControlFrame):
    """This frame is used to know when the internal pipelines have finished."""
    pass


@dataclass
class SyncNotifyFrame(ControlFrame):
    """This frame is used to know when notification has been received."""
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: data_frames.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x64\x61ta_frames.proto\x12\x0fpipeline_frames\"3\n\tTextFrame\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04text\x18\x03 \x01(\t\"y\n\rAudioRawFrame\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05\x61udio\x18\x03 \x01(\x0c\x12\x13\n\x0bsample_rate\x18\x04 \x01(\r\x12\x14\n\x0cnum_channels\x18\x05 \x01(\r\x12\x14\n\x0csample_width\x18\x06 \x01(\r\"d\n\rImageRawFrame\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05image\x18\x03 \x01(\x0c\x12\x0c\n\x04size\x18\x04 \x01(\t\x12\x0e\n\x06\x66ormat\x18\x05 \x01(\t\x12\x0c\n\x04mode\x18\x06 \x01(\t\"\x9e\x01\n\x05\x46rame\x12*\n\x04text\x18\x01 \x01(\x0b\x32\x1a.pipeline_frames.TextFrameH\x00\x12/\n\x05\x61udio\x18\x02 \x01(\x0b\x32\x1e.pipeline_frames.AudioRawFrameH\x00\x12/\n\x05image\x18\x03 \x01(\x0b\x32\x1e.pipeline_frames.ImageRawFrameH\x00\x42\x07\n\x05\x66rameb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'data_frames_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TEXTFRAME']._serialized_start=38
  _globals['_TEXTFRAME']._serialized_end=89
  _globals['_AUDIORAWFRAME']._serialized_start=91
  _globals['_AUDIORAWFRAME']._serialized_end=212
  _globals['_IMAGERAWFRAME']._serialized_start=214
  _globals['_IMAGERAWFRAME']._serialized_end=314
  _globals['_FRAME']._serialized_start=317
  _globals['_FRAME']._serialized_end=475
# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import asyncio
from typing import Awaitable, Callable, List

from apipeline.frames.sys_frames import Frame, SystemFrame
from apipeline.processors.async_frame_processor import AsyncFrameProcessor
from apipeline.processors.frame_processor import FrameDirection


class IdleFrameProcessor(AsyncFrameProcessor):
    """This class waits to receive any frame or list of desired frames within a
    given timeout. If the timeout is reached before receiving any of those
    frames the provided callback will be called.

    The callback can then be used to push frames downstream by using
    `queue_frame()` (or `push_frame()` for system frames).

    """

    def __init__(
            self,
            *,
            callback: Callable[["IdleFrameProcessor"], Awaitable[None]],
            timeout: float,
            types: List[type] = [],
            **kwargs):
        super().__init__(**kwargs)

        self._callback = callback
        self._timeout = timeout
        self._types = types

        self._create_idle_task()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, SystemFrame):
            await self.push_frame(frame, direction)
        else:
            await self.queue_frame(frame, direction)

        # If we are not waiting for any specific frame set the event, otherwise
        # check if we have received one of the desired frames.
        if not self._types:
            self._idle_event.set()
        else:
            for t in self._types:
                if isinstance(frame, t):
                    self._idle_event.set()

        # If we are not waiting for any specific frame set the event, otherwise
    async def cleanup(self):
        self._idle_task.cancel()
        await self._idle_task

    def _create_idle_task(self):
        self._idle_event = asyncio.Event()
        self._idle_task = self.get_event_loop().create_task(self._idle_task_handler())

    async def _idle_task_handler(self):
        while True:
            try:
                await asyncio.wait_for(self._idle_event.wait(), timeout=self._timeout)
            except asyncio.TimeoutError:
                await self._callback(self)
            except asyncio.CancelledError:
                break
            finally:
                self._idle_event.clear()
